##############################################################################
#
# Copyright (c) 2006 TINY SPRL. (http://tiny.be) All Rights Reserved.
#
# $Id: product_extended.py 3110 2006-05-09 12:11:24Z mvd $
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from osv import fields
from osv import osv
import ir

class product_supplierinfo(osv.osv):
	_inherit = 'product.supplierinfo'
	_name = "product.supplierinfo"

	def _last_order(self, cr, uid, ids, name, arg, context):
		res = {}
		for supinfo in self.browse(cr, uid, ids):
			cr.execute("select po.id, max(po.date_approve) from purchase_order as po, purchase_order_line as line where po.id=line.order_id and product_id=%d and partner_id=%d and state='approved' group by po.id", (supinfo.product_id.id, supinfo.name.id,))
			record = cr.fetchone()
			if record:
				res[supinfo.id] = record[0]
			else:
				res[supinfo.id] = False
		return res

	def _last_order_date(self, cr, uid, ids, name, arg, context):
		res = {}
		po = self.pool.get('purchase.order')
		last_orders = self._last_order(cr, uid, ids, name, arg, context)
		dates = po.read(cr, uid, filter(None, last_orders.values()), ['date_approve'])
		for suppinfo in ids:
			date_approve = [x['date_approve'] for x in dates if x['id']==last_orders[suppinfo]]
			if date_approve:
				res[suppinfo] = date_approve[0]
			else:
				res[suppinfo] = False
		return res

	_columns = {
		'last_order' : fields.function(_last_order, type='many2one', obj='purchase.order', method=True, string='Last Order'),
		'last_order_date' : fields.function(_last_order_date, type='date', method=True, string='Last Order date'),
	}
product_supplierinfo()


class product_product(osv.osv):
	_name = 'product.product'
	_inherit = 'product.product'
	
	def _running_po(self, cr, uid, ids, name, arg, context):
		res = {}
		for product in self.browse(cr, uid, ids):
			cr.execute("SELECT po.id, max(po.date_approve) FROM purchase_order as po, purchase_order_line as line \
						WHERE po.id=line.order_id AND product_id=%d AND state in ('done', 'cancel') GROUP BY po.id", (product.id,))
			record = cr.fetchone()
			if record:
				res[product.id] = record[0]
			else:
				res[product.id] = False
		return res

	_columns = {
		'running_purchase' : fields.function(_running_po, type='many2one', obj='purchase.order', method=True, string='Running order'),
		'calculate_price': fields.boolean('Compute price'),
	}

	_defaults = {
		'calculate_price': lambda w,x,y,z: False,
	}

	def compute_price(self, cr, uid, ids, *args):
		for prod_id in ids:
			bom_ids = osv.osv_pools.get('mrp.bom').search(cr, uid, [('product_id', '=', prod_id)])
			if bom_ids:
				for bom in osv.osv_pools.get('mrp.bom').browse(cr, uid, bom_ids):
					self._calc_price(cr, uid, bom)
		return True
					
	def _calc_price(self, cr, uid, bom):
		if not bom.product_id.calculate_price:
			return bom.product_id.standard_price
		else:
			price = 0
			if bom.bom_lines:
				for sbom in bom.bom_lines:
					price += self._calc_price(cr, uid, sbom) * sbom.product_qty
			else:
				pricelist_id = ir.ir_get(cr,uid,'meta','product.pricelist.purchase',[('res.partner',bom.product_id.seller_ids[0].name.id)])[0][2]
				price += osv.osv_pools.get('product.pricelist').price_get(cr, uid, [pricelist_id], bom.product_id.id, bom.product_qty, 'standard')[pricelist_id]
				
			if bom.routing_id:
				for wline in bom.routing_id.workcenter_lines:
					wc = wline.workcenter_id
					cycle = wline.cycle_nbr
					hour = (wc.time_start + wc.time_stop + cycle * wc.time_cycle) *  (wc.time_efficiency or 1.0)
					price += wc.costs_cycle * cycle + wc.costs_hour * hour
			self.write(cr, uid, [bom.product_id.id], {'standard_price' : price})
			return price
product_product()
